import {
    AppModel,
    DerivedData,
    DerivedFunctions,
} from "../generated_module/src/types/AutoModuleTypes";
import * as Processor from "@microchip/scf-common/lib/Processor";
import { ClockCalculations } from "./ClockCalculations";
import * as i2c_host_basic from "@microchip/i2c-host-basic";
import * as i2c_client_basic from "@microchip/i2c-client-basic";
import { Arguments, Interface } from "@microchip/scf-interface";
import { getGeneratorModel } from "./GeneratorModel";
import { I2CModeFunctions, getI2CModeFunctions } from "./I2CMode";
import * as init_system from "@microchip/initializer-system";
import { intf_args } from "@microchip/pin-standard/lib/pin-standard";
import { MyComponentNames } from "../generated_module/src/types";
import {
    getHeaderPathValidator,
    getCFunctionValidator,
} from "@microchip/scf-validators/lib";
import _find from "lodash.find";
import { ExportArg } from "@microchip/scf-automodule-interface/lib/src/InterfaceTypes";
import * as args_processor from "./ArgsProcessorHelper";
import { getCustomName } from "./GeneratorModel";

export const HOST_CALLBACK_EVENT = "i2cHostCallbackEvent";
export const CLIENT_CALLBACK_EVENT = "i2cClientCallbackEvent";
const DEFAULT_HOST_CALLBACK_NAME = "Host";
const DEFAULT_CLIENT_CALLBACK_NAME = "Client";

export const getDerivedData = (dataModel: AppModel): DerivedData => {
    if (dataModel) {
        return new MyDerivedData(dataModel);
    }
    return new EmptyDerivedData();
};

//This data will be used at the creator stage and relies on only the state
class EmptyDerivedData implements DerivedData {
    getMyFunctions = (): DerivedFunctions => {
        return {};
    };
    getModel = (): AppModel | undefined => {
        return undefined;
    };
}

//This data will be used at the processor stage
class MyDerivedData implements DerivedData {
    private dataModel: AppModel;
    private clockCalc: ClockCalculations;

    readonly interfacesInfo: args_processor.InterfacesInfo = {
        i2c_host_basic: {
            exportId: "i2c_basic",
            interfaceId: i2c_host_basic.Interface.getInterfaceId(),
        },
        i2c_client_basic: {
            exportId: "i2c_basic",
            interfaceId: i2c_client_basic.Interface.getInterfaceId(),
        },
    };

    constructor(dataModel: AppModel) {
        this.dataModel = dataModel;
    }

    private getCalc = (): ClockCalculations => {
        if (!this.clockCalc) {
            this.clockCalc = new ClockCalculations({
                clkPerInHz: Number(
                    this.dataModel.getImportValue("osc_clocks")?.fosc ?? 4000000,
                ),
                fSclInHz: this.dataModel.getComponentValue("desiredClockSpeed") ?? 0,
                tRiseInUs: 0.1,
            });
        }
        return this.clockCalc;
    };

    getMyFunctions = (): DerivedFunctions => {
        return {
            overrideDefaultValues: this.overrideDefaultValues,
            moduleName: this.getModuleName,
            componentName: this.componentName,
            mode: this.getRequestedMode,
            interruptDriven: this.interruptDriven,
            interruptDrivenVisible: (): boolean => true,
            isHostModeInterrupt: (): boolean => this.isHostModeInterrupt(),
            isHostModePolling: (): boolean => this.isHostModePolling(),
            isClientModeInterrupt: (): boolean => this.isClientModeInterrupt(),
            isClientModePolling: (): boolean => this.isClientModePolling(),
            isHostMode: (): boolean => this.isHostMode(),
            isClientMode: (): boolean => this.isClientMode(),
            "i2c-host-basic_payload": this.getHostPayload,
            "i2c-client-basic_payload": this.getClientPayload,
            enableMctrla: this.enableMctrla,
            enableSctrla: this.enableSctrla,
            mbaud: this.mbaud,
            showCommonOption: (): boolean => this.showCommonOption(),
            // saddr: (): number | undefined =>
            //     this.dataModel.getComponentValue("clientAddress"),
            isHostInterfaceUsed: this.isHostInterfaceLoaded,
            isClientInterfaceUsed: this.isClientInterfaceLoaded,
            saddr: this.getClientAddress,
            addrmaskSaddrmask: this.getClientMask,
            actualClockSpeed: this.actualClockSpeed,
            pinData: this.getMyPinData,
            templateData: (): any => getGeneratorModel(this.dataModel),
            desiredClockSpeed: this.desiredClockSpeed,
            desiredClockSpeedValidator: this.desiredClockSpeedValidator,
            slaveAddressValidator: this.slaveAddressValidator,
            alerts: this.getMyAlerts,
            initializer_system_results: this.getInitializerSystemResults,
            interruptComponentName: this.interruptComponentName,
            pin_standard_args: this.getPinstandardArgs,
            filterInterrupts: this.filterInterrupts,
            rienMctrla: (): boolean | undefined =>
                this.getModel().isComponentValue("interruptDriven") ? undefined : false,
            wienMctrla: (): boolean | undefined =>
                this.getModel().isComponentValue("interruptDriven") ? undefined : false,
            apienSctrla: (): boolean | undefined =>
                this.getModel().isComponentValue("interruptDriven") ? undefined : false,
            dienSctrla: (): boolean | undefined =>
                this.getModel().isComponentValue("interruptDriven") ? undefined : false,
            pienSctrla: (): boolean | undefined =>
                this.getModel().isComponentValue("interruptDriven") ? undefined : false,
            getActiveName: this.getActiveName,
        };
    };

    private getActiveName = (): string => {
        return this.dataModel.getHardware()?.getName() + "_Peripheral";
    };

    // private getModuleGroupName = (): string => {
    //     return "i2c";
    // };

    private filterInterrupts = (interruptName: string): boolean => {
        switch (interruptName) {
            case "WIE":
                return this.getModel().isComponentValue("interruptDriven");
            case "RIE":
                return this.getModel().isComponentValue("interruptDriven");
            case "APIE":
                return this.getModel().isComponentValue("interruptDriven");
            case "DIE":
                return this.getModel().isComponentValue("interruptDriven");
            case "PIE":
                return this.getModel().isComponentValue("interruptDriven");
            default:
                return false;
        }
    };

    private getPinstandardArgs = (): intf_args => {
        const args: intf_args = {
            queryAdapters: ["portmux"],
        };
        return args;
    };

    private overrideDefaultValues = (componentName: string): any => {
        switch (componentName) {
            case "componentName":
                return (
                    this.dataModel.getHardware()?.getName().replace("TWI", "I2C") ?? ""
                );
        }
    };

    public getModel = (): AppModel => {
        return this.dataModel;
    };

    // public EnableClientMode = (): string => {

    //     if (this.isClientInterfaceLoaded()) {
    //         return "enabled";
    //     }
    //     return "disabled";
    // };

    public isClientInterfaceLoaded = (): boolean => {
        const exportData = this.dataModel.getExportInterfaces().getInterfaceData();
        const clientData = _find(
            exportData,
            (i) => i.interface === i2c_client_basic.Interface.getInterfaceId().name,
        );

        if (clientData) {
            return true;
        }
        return false;
    };

    public isHostInterfaceLoaded = (): boolean => {
        const exportData = this.dataModel.getExportInterfaces().getInterfaceData();
        const hostData = _find(
            exportData,
            (i) => i.interface === i2c_host_basic.Interface.getInterfaceId().name,
        );

        if (hostData) {
            return true;
        }
        return false;
    };

    private getHostExportArg = (componentName: string): ExportArg | undefined => {
        const exportData = this.getModel().getExportInterfaces().getInterfaceData();

        const expArg = _find(
            exportData,
            (i) =>
                i.interface === i2c_host_basic.Interface.getInterfaceId().name &&
                i.name === componentName,
        );
        return expArg;
    };

    private getClientExportArg = (componentName: string): ExportArg | undefined => {
        const exportData = this.getModel().getExportInterfaces().getInterfaceData();

        const expArg = _find(
            exportData,
            (i) =>
                i.interface === i2c_client_basic.Interface.getInterfaceId().name &&
                i.name === componentName,
        );
        return expArg;
    };

    // private getRequestedArg = (argName: string): any => {
    //     if (this.isHostInterfaceLoaded()) {
    //         return this.getHostExportArg(argName);
    //     } else if (this.isClientInterfaceLoaded()) {
    //         return this.getClientExportArg(argName);
    //     }
    //     return undefined;
    // };

    private getModeFunctions = (): I2CModeFunctions | undefined => {
        if (this.dataModel) {
            return getI2CModeFunctions(this.dataModel);
        }
        return undefined;
    };

    public getModuleName = (): string => {
        return this.dataModel.getHardware()?.getName().toUpperCase() ?? "";
    };

    private getRequestedMode = (): string | undefined => {
        if (this.isHostInterfaceLoaded()) {
            return "Host";
        } else if (this.isClientInterfaceLoaded()) {
            return "Client";
        } else {
            return undefined;
        }
    };

    private componentName = (): string | undefined => {
        // return (
        //     // this.getFirstExportArg<i2c_host_basic.Arguments>(i2c_host_basic.Interface)
        //     //     ?.customName ??
        //     // this.getFirstExportArg<i2c_client_basic.Arguments>(i2c_client_basic.Interface)
        //     //     ?.customName
        //     this.getCustomName()
        // );
        let customName = "";
        const i2cHostInterface = i2c_host_basic.Interface.getInterfaceId();
        const i2cClientInterface = i2c_client_basic.Interface.getInterfaceId();
        if (this.getInterfaceUsage(i2cHostInterface.name)) {
            customName = getCustomName(this.dataModel, i2cHostInterface);
        } else if (this.getInterfaceUsage(i2cClientInterface.name)) {
            customName = getCustomName(this.dataModel, i2cClientInterface);
        }
        if (customName === "") {
            customName =
                // this.dataModel.getHardware()?.getName().toLocaleUpperCase() ?? "";
                this.dataModel.getHardware()?.getName().replace("TWI", "I2C") ?? "I2C";
        }
        return customName;
    };
    public getInterfaceUsage = (interfaceName: string): boolean => {
        const exportData:
            | ExportArg[]
            | undefined = this.dataModel.getExportInterfaces().getInterfaceData();
        const interfaceData = _find(exportData, (i) => i.interface === interfaceName);
        if (interfaceData) {
            return true;
        }
        return false;
    };

    private interruptDriven = (): boolean | undefined => {
        return (
            this.getFirstExportArg<i2c_host_basic.Arguments>(i2c_host_basic.Interface)
                ?.interruptDriven ??
            this.getFirstExportArg<i2c_client_basic.Arguments>(i2c_client_basic.Interface)
                ?.interruptDriven
        );
    };

    private desiredClockSpeed = (): number | undefined => {
        const desiredClockSpeed = this.getFirstExportArg<i2c_host_basic.Arguments>(
            i2c_host_basic.Interface,
        )?.requestedSpeed_Hz;
        return desiredClockSpeed ? Number(desiredClockSpeed) : undefined;
    };

    // private interruptsEnable = (): boolean => {
    //     return this.getRequestedArg("interruptDriven");
    // };

    private getFirstExportArg = <T extends Arguments>(
        id: Interface<T>,
    ): T | undefined => {
        const i2cExport = this.dataModel
            .getExportInterfaces()
            .getInterface("i2c_basic", id.getInterfaceId());
        return i2cExport?.args?.[Object.keys(i2cExport.args)[0]];
    };

    private getHostPayload = (): i2c_host_basic.ProcessedPayload => {
        const headerFiles = i2c_host_basic.Interface.createFirmwareApi(
            this.dataModel.getName(),
        ).headerFiles;

        // const modInstance = this.dataModel.getHardware()?.getPeripheral()?.instance;
        // let modInstanceNo = 1; //instance cannot be < 1
        // if (modInstance) {
        //     modInstanceNo = Number(modInstance);
        // }
        const modInstance: string =
            this.dataModel.getHardware()?.getPeripheral()?.instance ?? "";
        let modInstanceNo = 0;
        if (modInstance) {
            modInstanceNo = Number(modInstance);
        }
        return {
            interfaceApi: i2c_host_basic.Interface.createFirmwareApi(
                this.dataModel.getComponentValue("componentName"),
                headerFiles,
            ),
            moduleInstance: modInstanceNo,
            moduleName: this.dataModel.getHardware()?.getPeripheral()?.name ?? "",
            hasInterrupts: true,
            interruptDriven: this.dataModel.isComponentValue("interruptDriven"),
            calculatedSpeed_Hz: String(
                this.dataModel.getComponentValue("actualClockSpeed") ?? "0",
            ),
            i2cHostCallbackEvent: this.getPayloadCallbackData(HOST_CALLBACK_EVENT),
        };
    };

    private getClientPayload = (): i2c_client_basic.ProcessedPayload => {
        const headerFiles = i2c_client_basic.Interface.createFirmwareApi(
            this.dataModel.getName(),
        ).headerFiles;
        const modInstance = this.dataModel.getHardware()?.getPeripheral()?.instance;
        let modInstanceNo = 0;
        if (modInstance) {
            modInstanceNo = Number(modInstance);
        }
        return {
            interfaceApi: i2c_client_basic.Interface.createFirmwareApi(
                this.dataModel.getComponentValue("componentName"),
                headerFiles,
            ),
            moduleInstance: modInstanceNo,
            moduleName: this.dataModel.getHardware()?.getPeripheral()?.name ?? "",
            hasInterrupts: true,
            interruptDriven: this.dataModel.isComponentValue("interruptDriven"),
            clockStretching: false,
            clientAddress: Number(this.dataModel.getComponentValue("clientAddress")),
            clientMask: Number(this.dataModel.getComponentValue("addrmaskSaddrmask")),
            i2cClientCallbackEvent: this.getPayloadCallbackData(CLIENT_CALLBACK_EVENT),
            isClockStrechAlwaysEnabled: true,
            maxClientAddress: 127,
            maxClientMask: 127,
            minClientAddress: 0,
        };
    };

    private isValidPath = (headerPath: string): boolean => {
        const pathValidator = getHeaderPathValidator()
            .getRjsfValidation()
            .pattern.toString();
        const pathRegex = new RegExp(pathValidator);
        return pathRegex.test(headerPath);
    };

    private isValidCallbackName = (callbackName: string): boolean => {
        const cValiator = getCFunctionValidator().getRjsfValidation().pattern.toString();
        const nameRegex = new RegExp(cValiator);
        return nameRegex.test(callbackName);
    };

    public getDefaultCallbackName = (callback: string): string => {
        switch (callback) {
            case CLIENT_CALLBACK_EVENT:
                return DEFAULT_CLIENT_CALLBACK_NAME;
            case HOST_CALLBACK_EVENT:
                return DEFAULT_HOST_CALLBACK_NAME;
            default:
                return "";
        }
    };

    public getPayloadCallbackData = (callback: string): any => {
        let cbData = this.getHostExportArg(callback)?.value ?? "";

        if (callback === CLIENT_CALLBACK_EVENT) {
            cbData = this.getClientExportArg(callback)?.value ?? "";
        }

        const defaultCbName = this.getDefaultCallbackName(callback);
        const defultHeaderPath =
            this.dataModel.getHardware()?.getPeripheral()?.name.toLocaleLowerCase() +
            ".h";
        if (cbData != "") {
            let cbName = cbData["name"];
            let headerPath = cbData["header"];
            if (!this.isValidCallbackName(cbName)) {
                cbName = defaultCbName;
            }
            if (!this.isValidPath(headerPath)) {
                headerPath = defultHeaderPath;
            }
            cbData = {
                ...cbData,
                name: cbName,
                header: headerPath,
            };
        } else {
            cbData = {
                name: defaultCbName,
                header: defultHeaderPath,
            };
        }
        return cbData;
    };

    private getInitializerSystemResults = (): init_system.Results => {
        const api = getGeneratorModel(this.dataModel);
        const header = api.header;
        return {
            initializer: {
                name: api.api.Initialize?.name,
                header: {
                    name: header?.name,
                    path: header?.path?.replace("../", ""),
                },
            },
        };
    };

    private interruptComponentName = (
        interruptName: string,
    ): MyComponentNames | undefined => {
        if (interruptName === "APIE") {
            return "apienSctrla";
        } else if (interruptName === "DIE") {
            return "dienSctrla";
        } else if (interruptName === "PIE") {
            return "pienSctrla";
        } else if (interruptName === "RIE") {
            return "rienMctrla";
        } else if (interruptName === "WIE") {
            return "wienMctrla";
        } else {
            return undefined;
        }
    };

    private isHostMode = (): boolean => {
        return this.dataModel.getComponentValue("mode") === "Host";
    };

    private isClientMode = (): boolean => {
        return this.dataModel.getComponentValue("mode") === "Client";
    };

    private isHostModeInterrupt = (): boolean => {
        return (
            this.dataModel.isComponentValue("interruptDriven") &&
            this.dataModel.getComponentValue("mode") === "Host"
        );
    };

    private isHostModePolling = (): boolean => {
        return (
            !this.dataModel.isComponentValue("interruptDriven") &&
            this.dataModel.getComponentValue("mode") === "Host"
        );
    };

    private isClientModeInterrupt = (): boolean => {
        return (
            this.dataModel.isComponentValue("interruptDriven") &&
            this.dataModel.getComponentValue("mode") === "Client"
        );
    };

    private isClientModePolling = (): boolean => {
        return (
            !this.dataModel.isComponentValue("interruptDriven") &&
            this.dataModel.getComponentValue("mode") === "Client"
        );
    };

    private enableMctrla = (): string | undefined => {
        if (this.isHostInterfaceLoaded()) {
            return this.boolToReg(this.isHostInterfaceLoaded());
        } else {
            return this.boolToReg(
                this.dataModel.isComponentValue("enableI2C") &&
                    this.dataModel.getComponentValue("mode") === "Host",
            );
        }
    };

    private enableSctrla = (): string | undefined => {
        if (this.isClientInterfaceLoaded()) {
            return this.boolToReg(this.isClientInterfaceLoaded());
        } else {
            return this.boolToReg(
                this.dataModel.isComponentValue("enableI2C") &&
                    this.dataModel.getComponentValue("mode") === "Client",
            );
        }
    };

    private getClientAddress = (): string | undefined => {
        if (this.isClientInterfaceLoaded()) {
            const address = Number(this.getClientExportArg("clientAddress")?.value);
            if (address) {
                return address.toString();
            }
        } else {
            if (this.dataModel.getComponentValue("mode") === "Client") {
                let saddr: number | undefined =
                    this.dataModel.getComponentValue("clientAddress") ?? 0;
                if (saddr != undefined) {
                    saddr = (saddr << 0) & 0xff;
                    return saddr.toString();
                }
            }
        }
    };

    private getClientMask = (): string | undefined => {
        const mask = Number(this.getClientExportArg("clientMask")?.value);
        if (mask) {
            return mask.toString();
        }
    };

    private mbaud = (): number => {
        return this.getCalc().getMBAUD();
    };

    private actualClockSpeed = (): number => {
        return this.getCalc().getActualClockSpeed();
    };

    private getMyPinData = (): any => {
        return {};
    };

    private getMyAlerts = (): Processor.Alert[] => {
        const alerts: Processor.Alert[] = [];

        alerts.push({
            text: "Dual Mode driver is not supported in MCC.",
            type: Processor.AlertTypes.Hint,
        });

        if (
            this.dataModel.isComponentValue("fmpenCtrla") &&
            Number(this.dataModel.getImportValue("osc_clocks")?.fosc ?? 0) < 16e6
        ) {
            alerts.push({
                text:
                    "FMP mode is selected.The Peripheral Clock (CLK_PER) has to run at 16MHz or faster.",
                type: Processor.AlertTypes.Warning,
            });
        }

        if (this.isHostInterfaceLoaded()) {
            // const reqClock = Number(this?.getHostExportArg("desiredClockSpeed")?.value);
            const reqClock = Number(
                this.dataModel.getComponentValue("desiredClockSpeed"),
            );
            if (reqClock != 100000 && reqClock != 400000) {
                alerts.push({
                    text:
                        " For requested " +
                        reqClock +
                        " I2C clock speed, Please change the clock source to More than 16MHz under the CLKCTRL. ",
                    type: Processor.AlertTypes.Hint,
                });
            }
        }

        alerts.push(...(this.getModeFunctions()?.getAlerts() ?? []));

        return alerts;
    };
    // private isAssignedExport = (key: MyExportNames, id: Interface): boolean => {
    //     const exportInterface = this.dataModel
    //         .getExportInterfaces()
    //         .getInterface(key, id.getInterfaceId());
    //     return exportInterface?.args != undefined;
    // };

    private desiredClockSpeedValidator = (): any => {
        return {
            minimum: 1,
            maximum: Number(this.dataModel.getImportValue("osc_clocks")?.fosc) ?? 1000000,
            multipleOf: 1,
        };
    };

    private slaveAddressValidator = (): any => {
        return {
            minimum: 0,
            maximum: 127,
            multipleOf: 1,
        };
    };

    private boolToReg = (enable: boolean): string | undefined => {
        if (enable != undefined) {
            return enable ? "enabled" : "disabled";
        }
    };

    private showCommonOption = (): boolean => {
        if (this.isHostInterfaceLoaded() || this.isClientInterfaceLoaded()) {
            return false;
        }
        return true;
    };
}
