import * as i2c_client_interface from "@microchip/i2c-client-basic";
import * as i2c_host_interface from "@microchip/i2c-host-basic";
import { HeaderFile } from "@microchip/scf-interface";
import { AppModel } from "@microchip/scf-automodule-interface/lib/src/types";
import { InterfaceId } from "@microchip/scf-common/lib/InterfaceId";
import { getCFunctionValidator } from "@microchip/scf-validators/lib";

export const getGeneratorModel = (model: AppModel): any => {
    const interrupts = model.getHardware()?.getPeripheral().interrupts;
    const registers = model.getRegisterValues();
    const moduleName = model.getName();
    const componentName = model.getComponentValue("componentName");

    const i2cInterface =
        model.getComponentValue("mode") === "Client"
            ? i2c_client_interface.Interface.createFirmwareApi(
                  componentName,
                  getHeaderFiles(model),
              )
            : i2c_host_interface.Interface.createFirmwareApi(
                  componentName,
                  getHeaderFiles(model),
              );

    const getFoscValue = (): string => {
        const fosc = Number(model?.getImportValue("osc_clocks")?.fosc ?? 0);
        return fosc.toString();
    };

    return {
        productName: "",
        productVersion: "",
        selectedDevice: "",
        compiler: "XC8 v2.20 and above",
        tool: "MPLABX v5.40  and above",
        hostDriverVersion: "2.0.0",
        clientDriverVersion: "2.0.0",

        moduleNameLowerCase: moduleName.toLowerCase(),
        moduleNameUpperCase: moduleName.toUpperCase(),
        moduleGroupNameLowerCase: getModuleGroupName().toLowerCase(),
        moduleGroupNameUpperCase: getModuleGroupName().toUpperCase(),
        moduleGroupNameLowerCase1: getModuleGroupName1().toLowerCase(),
        moduleGroupNameUpperCase1: getModuleGroupName1().toUpperCase(),
        moduleInstance: moduleName,
        componentName: model.getComponentValue("componentName"),
        componentNameLowerCase: componentName.toLowerCase(),
        componentNameUpperCase: componentName.toUpperCase(),
        header: i2cInterface?.headerFiles?.[0],
        api: i2cInterface.api,
        simpleApi: i2cInterface.simpleApi,
        interface: i2cInterface,
        instanceNumber: model.getPeripheralDescription()?.instance,
        isInterruptDriven: model.getComponentValue("interruptDriven") ?? false,

        CLK_PER: getFoscValue(),
        SCL_Frequency: String(model.getComponentValue("actualClockSpeed")),
        Trise: 0.1,

        //Registers and Settings
        initRegisters: registers,
        initInterrupts: interrupts,

        ...registers,
        ...interrupts,
    };
};

export const getHeaderFiles = (model: AppModel): HeaderFile[] => {
    return [
        {
            name: `${getFilePrefix(model)}.h`,
            path:
                model.getComponentValue("mode") === "Host" ? "i2c_host/" : "i2c_client/",
        },
    ];
};

const getFilePrefix = (model: AppModel): string => {
    return (model.getHardware()?.getName() ?? "I2C").toLowerCase();
};

const getModuleGroupName = (): string => {
    return "I2C";
};

const getModuleGroupName1 = (): string => {
    return "TWI";
};

export const getCustomName = (dataModel: AppModel, interfaceId: InterfaceId): string => {
    const data = dataModel

        ?.getExportInterfaces()

        .getInterface("i2c_basic", interfaceId)?.results;

    let customNameStr = ""; //Default

    if (data) {
        for (const moduleId in data) {
            const customNameData = data[moduleId];

            if (customNameData?.customName) {
                customNameStr = customNameData["customName"] ?? "";
            }
        }
    }

    if (!isValidCustomName(customNameStr)) {
        customNameStr = dataModel.getName();
    }

    return customNameStr;
};

//TO DO: Ideally the customName coming from the interface needs to be validated
//at the interface level itself
export const isValidCustomName = (customName: string): boolean => {
    if (customName != undefined) {
        const cValiator = getCFunctionValidator().getRjsfValidation().pattern.toString();
        const nameRegex = new RegExp(cValiator);
        return nameRegex.test(customName);
    }
    return false;
};
